import os
import csv
import json
import math
import time
import copy
import random
import string
import datetime

GOLD_FILE_IN = "/Users/Daniel/Desktop/TDDD82/alg_eval/GOLD_courtyard_2000.json"
GOLD_FILE_OUT = "/Users/Daniel/Desktop/TDDD82/alg_eval/GOLD_courtyard_2000.csv"

FOLDER_IN = "/Users/Daniel/Desktop/TDDD82/alg_eval/courtyard_tests"
FOLDER_OUT = "/Users/Daniel/Desktop/TDDD82/alg_eval/out/" + \
    datetime.datetime.now().isoformat()

os.makedirs(FOLDER_OUT)

FILE = ""

V_FOV = 60
H_FOV = math.degrees(2 * math.atan(math.tan(math.radians(V_FOV)/2) * (1920/1080)))

full_path_in = os.path.join(FOLDER_IN, FILE + ".json")
full_path_out = os.path.join(FOLDER_OUT, FILE + ".csv")

full_path_summary = os.path.join(FOLDER_OUT, "summary.csv")

CSV_SETTINGS = dict(delimiter=',', quotechar='$', quoting=csv.QUOTE_MINIMAL)


def parse_file(file_path, gold=False):
    with open(file_path) as _file:
        content = _file.read()

    print("Parsing...", end="\r")
    total_lines = len(content.splitlines())

    if content.startswith("["):
        return json.loads(content)

    new_content = []
    for i, line in enumerate(content.splitlines()):
        print("Parsing... line %s of %s" % (i, total_lines), end="\r")
        new_content.append(line + ("," if i != total_lines-1 else ""))

    content = "\n".join(new_content)
    content = "[%s]" % content
    print("Parsing complete" + " "*20)

    # with open(file_path, "w") as _file:
    #     _file.write(content)

    print("Loading JSON...", end="\r")
    json_content = json.loads(content)

    # with open(file_path, "w") as _file:
    #     _file.write(json.dumps(json_content, indent=2))
    print("Loading JSON complete")

    print("Sorting list...", end="\r")
    json_content.sort(key=lambda entry: entry["eulerAngles"]["y"])
    print("Sorting list complete")

    t = time.time()

    gold_list = []

    if gold:
        for i, gold_row in enumerate(json_content):
            neighbors = 0

            angle = gold_row["eulerAngles"]["y"]
            objects = []
            print("Normalizing gold... %s/%s" % (i+1, len(json_content)), end="\r")

            for inner_gold_row in json_content:
                inner_angle = inner_gold_row["eulerAngles"]["y"]

                if abs(angle - inner_angle) < H_FOV/2 or \
                    abs(360 + (angle - inner_angle)) < H_FOV/2 or \
                        abs(360 - (angle - inner_angle)) < H_FOV/2:
                    neighbors += 1

                    already_added = [_o["name"] for _o in objects]

                    for o in inner_gold_row["visibleObjects"]:
                        if not o["name"] in already_added:
                            objects.append(o)

            gold_list.append(objects)

        for i, gold_objects in enumerate(gold_list):
            json_content[i]["visibleObjects"] = gold_objects

        delta = time.time() - t
        print("Normalizing gold complete, took: %s:%.0f" % (math.floor(delta/60), delta % 60))

    with open(file_path, "w") as _file:
        _file.write(json.dumps(json_content, indent=2))

    return json_content


def get_closest_gold(gold, sample_point):
    lowest_diff = 360
    closest_gold = gold[0]
    sample_angle = sample_point["eulerAngles"]["y"]

    for gold_row in gold:
        diff = abs(gold_row["eulerAngles"]["y"] - sample_angle)
        if diff < lowest_diff:
            closest_gold = gold_row
            lowest_diff = diff
        else:
            break

    return closest_gold


def find_object_data_by_name(data_point, name):
    return next(iter([o for o in data_point["visibleObjects"]
                if o["name"] == name]), None)


def calculate_false_negatives(closest_gold, sample_point):
    true_negatives = 0

    for object_data in closest_gold["visibleObjects"]:
        if find_object_data_by_name(sample_point, object_data["name"]) is None:
            true_negatives += 1

    return true_negatives


def calculate_false_positives(closest_gold, sample_point):
    false_positives = 0

    for object_data in sample_point["visibleObjects"]:
        if find_object_data_by_name(closest_gold, object_data["name"]) is None:
            false_positives += 1

    return false_positives


def calculate_true_positives(closest_gold, sample_point):
    true_positives = 0

    for object_data in closest_gold["visibleObjects"]:
        if find_object_data_by_name(sample_point, object_data["name"]) is not None:
            true_positives += 1

    return true_positives


def calculate_precision(true_positives, false_positives):
    return true_positives / (true_positives + false_positives)


def calculate_recall(true_positives, false_negatives):
    return true_positives / (true_positives + false_negatives)


def calculate_f1(precision, recall):
    return 2 * (precision * recall) / (precision + recall)


def to_csv(file_path, json_content, gold=None):
    total_rows = len(json_content)
    rows_written = 0
    
    total_objects = 0

    total_precision = 0
    total_recall = 0
    total_f1 = 0

    total_fps = 0

    with open(file_path, "w", newline="") as csv_file:
        writer = csv.writer(csv_file, **CSV_SETTINGS)

        if not gold:
            writer.writerow(["Deg", "Objects"])

            for gold_point in json_content:
                writer.writerow([
                    gold_point["eulerAngles"]["y"],
                    len(gold_point["visibleObjects"])
                ])
                rows_written += 1
                print("Converting to CSV... %.0f %%" %
                      ((rows_written / total_rows) * 100), end="\r")
        else:
            writer.writerow([
                "Deg",
                "Objects",
                "False negatives",
                "False positives",
                "True positives",
                "Precision",
                "Recall",
                "F1"
            ])

            for sample_point in json_content:
                angle = sample_point["eulerAngles"]["y"]

                if angle < 0:
                    continue

                closest_gold = get_closest_gold(gold, sample_point)
                
                false_negatives = calculate_false_negatives(closest_gold,
                                                            sample_point)
                false_positives = calculate_false_positives(closest_gold,
                                                            sample_point)
                true_positives = calculate_true_positives(closest_gold,
                                                          sample_point)

                objects = len(sample_point["visibleObjects"])

                precision = calculate_precision(true_positives, false_positives)
                recall = calculate_recall(true_positives, false_negatives)
                f1 = calculate_f1(precision, recall)
                
                total_objects += objects

                total_precision += precision
                total_recall += recall
                total_f1 += f1

                total_fps += sample_point["fps"]

                writer.writerow([
                    sample_point["eulerAngles"]["y"],
                    objects,
                    false_negatives,
                    false_positives,
                    true_positives,
                    precision,
                    recall,
                    f1
                ])

                rows_written += 1

                print("Converting to CSV... %.0f %%" %
                      ((rows_written / total_rows) * 100), end="\r")

    print("Converting to CSV complete")

    if gold:
        average_objects = total_objects / rows_written
        
        average_precision = total_precision / rows_written
        average_recall = total_recall / rows_written
        average_f1 = total_f1 / rows_written

        average_fps = total_fps / rows_written

        print("Objects: %s" % average_objects)
        print("Precision: %s" % average_precision)
        print("Recall: %s" % average_recall)
        print("F1: %s" % average_f1)
        print("FPS: %s" % average_fps)

        if not os.path.isfile(full_path_summary):
            with open(full_path_summary, "w") as csv_file:
                writer = csv.writer(csv_file, **CSV_SETTINGS)
                writer.writerow([
                    "Name",
                    "Average objects",
                    "Average precision",
                    "Average recall",
                    "Average F1",
                    "Average FPS"
                ])

        with open(full_path_summary, "a") as summary_csv:
            writer = csv.writer(summary_csv, **CSV_SETTINGS)
            writer.writerow([
                "".join(file_path.split(os.path.sep)[-1].split(".")[0:-1]),
                average_objects,
                average_precision,
                average_recall,
                average_f1,
                average_fps
            ])

json_gold = parse_file(GOLD_FILE_IN, gold=True)
to_csv(GOLD_FILE_OUT, json_gold)

if FILE:
    test_json = parse_file(full_path_in)
    to_csv(full_path_out, test_json, gold=json_gold)

else:
    print("Files will be saved to %s" % FOLDER_OUT)
    
    for file in os.listdir(FOLDER_IN):
        if file.split(os.path.sep)[-1].startswith("."): continue
        if "gold" in file:
            continue
        print("===========================")
        print("File: %s" % file)
        full_path_in = os.path.join(FOLDER_IN, file)
        full_path_out = os.path.join(FOLDER_OUT, file.replace(".json", ".csv"))
        test_json = parse_file(full_path_in)
        to_csv(full_path_out, test_json, gold=json_gold)
